import {
    SEMANTIC_ATTR8, SEMANTIC_ATTR9, SEMANTIC_ATTR10, SEMANTIC_ATTR11, SEMANTIC_ATTR12, SEMANTIC_ATTR13, SEMANTIC_ATTR14, SEMANTIC_ATTR15,
    SEMANTIC_BLENDINDICES, SEMANTIC_BLENDWEIGHT, SEMANTIC_COLOR, SEMANTIC_NORMAL, SEMANTIC_POSITION, SEMANTIC_TANGENT,
    SEMANTIC_TEXCOORD0, SEMANTIC_TEXCOORD1,
    SHADERTAG_MATERIAL,
    PIXELFORMAT_R8_G8_B8_A8
} from '../../constants.js';
import { shaderChunks } from '../chunks/chunks.js';
import { ChunkUtils } from '../chunk-utils.js';

import {
    BLEND_ADDITIVEALPHA, BLEND_NORMAL, BLEND_PREMULTIPLIED,
    FRESNEL_SCHLICK,
    LIGHTFALLOFF_LINEAR,
    LIGHTSHAPE_PUNCTUAL, LIGHTSHAPE_RECT, LIGHTSHAPE_DISK, LIGHTSHAPE_SPHERE,
    LIGHTTYPE_DIRECTIONAL, LIGHTTYPE_OMNI, LIGHTTYPE_SPOT,
    SHADER_DEPTH, SHADER_PICK,
    SHADOW_PCF3, SHADOW_PCF5, SHADOW_VSM8, SHADOW_VSM16, SHADOW_VSM32,
    SPECOCC_AO, SPECOCC_GLOSSDEPENDENT,
    SPECULAR_PHONG,
    SPRITE_RENDERMODE_SLICED, SPRITE_RENDERMODE_TILED, shadowTypeToString
} from '../../../scene/constants.js';
import { LightsBuffer } from '../../../scene/lighting/lights-buffer.js';
import { ShaderPass } from '../../../scene/shader-pass.js';

import { begin, end, fogCode, gammaCode, precisionCode, skinCode, tonemapCode, versionCode } from './common.js';
import { validateUserChunks } from '../chunks/chunk-validation.js';

const builtinAttributes = {
    vertex_normal: SEMANTIC_NORMAL,
    vertex_tangent: SEMANTIC_TANGENT,
    vertex_texCoord0: SEMANTIC_TEXCOORD0,
    vertex_texCoord1: SEMANTIC_TEXCOORD1,
    vertex_color: SEMANTIC_COLOR,
    vertex_boneWeights: SEMANTIC_BLENDWEIGHT,
    vertex_boneIndices: SEMANTIC_BLENDINDICES
};

const builtinVaryings = {
    vVertexColor: "vec4",
    vPositionW: "vec3",
    vNormalV: "vec3",
    vNormalW: "vec3",
    vTangentW: "vec3",
    vBinormalW: "vec3",
    vObjectSpaceUpW: "vec3",
    vUv0: "vec2",
    vUv1: "vec2"
};

class LitShader {
    constructor(device, options) {
        this.device = device;
        this.options = options;

        // resolve custom chunk attributes
        this.attributes = {
            vertex_position: SEMANTIC_POSITION
        };

        if (options.chunks) {
            this.chunks = {};

            const userChunks = options.chunks;

            // #if _DEBUG
            validateUserChunks(options.chunks);
            // #endif

            for (const chunkName in shaderChunks) {
                if (userChunks.hasOwnProperty(chunkName)) {
                    const chunk = userChunks[chunkName];
                    for (const a in builtinAttributes) {
                        if (builtinAttributes.hasOwnProperty(a) && chunk.indexOf(a) >= 0) {
                            this.attributes[a] = builtinAttributes[a];
                        }
                    }
                    this.chunks[chunkName] = chunk;
                } else {
                    this.chunks[chunkName] = shaderChunks[chunkName];
                }
            }
        } else {
            this.chunks = shaderChunks;
        }

        if (options.forceFragmentPrecision &&
            options.forceFragmentPrecision !== "highp" &&
            options.forceFragmentPrecision !== "mediump" &&
            options.forceFragmentPrecision !== "lowp") {
            options.forceFragmentPrecision = null;
        }

        if (options.forceFragmentPrecision) {
            if (options.forceFragmentPrecision === "highp" && device.maxPrecision !== "highp") {
                options.forceFragmentPrecision = "mediump";
            }
            if (options.forceFragmentPrecision === "mediump" && device.maxPrecision === "lowp") {
                options.forceFragmentPrecision = "lowp";
            }
        }

        this.lighting = (options.lights.length > 0) || !!options.dirLightMap || !!options.clusteredLightingEnabled;
        this.reflections = !!options.reflectionSource;
        if (!options.useSpecular) options.specularMap = options.glossMap = null;
        this.shadowPass = ShaderPass.isShadow(options.pass);
        this.needsNormal = this.lighting || this.reflections || options.useSpecular || options.ambientSH || options.heightMap || options.enableGGXSpecular ||
                            (options.clusteredLightingEnabled && !this.shadowPass) || options.clearCoatNormalMap;

        // generated by vshader
        this.varyings = "";
        this.vshader = null;

        // supplied by caller
        this.frontendDecl = null;
        this.frontendCode = null;
        this.frontendFunc = null;
        this.lightingUv = null;

        // generated by fshader
        this.fshader = null;
    }

    _getPassDefineString(pass) {
        if (pass === SHADER_PICK) {
            return '#define PICK_PASS\n';
        } else if (pass === SHADER_DEPTH) {
            return '#define DEPTH_PASS\n';
        } else if (ShaderPass.isShadow(pass)) {
            return '#define SHADOW_PASS\n';
        }
        return '';
    }

    _vsAddBaseCode(code, chunks, options) {
        code += chunks.baseVS;
        if (options.nineSlicedMode === SPRITE_RENDERMODE_SLICED ||
            options.nineSlicedMode === SPRITE_RENDERMODE_TILED) {
            code += chunks.baseNineSlicedVS;
        }
        return code;
    }

    _vsAddTransformCode(code, device, chunks, options) {
        code += this.chunks.transformVS;
        return code;
    }

    _setMapTransform(codes, name, id, uv) {
        const checkId = id + uv * 100;
        if (!codes[3][checkId]) {
            // upload a 3x2 matrix and manually perform the multiplication
            const varName = `texture_${name}MapTransform`;
            codes[0] += `uniform vec3 ${varName}0;\n`;
            codes[0] += `uniform vec3 ${varName}1;\n`;
            codes[1] += `varying vec2 vUV${uv}_${id};\n`;
            codes[2] += `   vUV${uv}_${id} = vec2(dot(vec3(uv${uv}, 1), ${varName}0), dot(vec3(uv${uv}, 1), ${varName}1));\n`;
            codes[3][checkId] = true;
        }
        return codes;
    }

    // Add "Base" Code section to fragment shader.
    _fsGetBaseCode() {
        const options = this.options;
        const chunks = this.chunks;
        let result = this.chunks.basePS;
        if (options.nineSlicedMode === SPRITE_RENDERMODE_SLICED) {
            result += chunks.baseNineSlicedPS;
        } else if (options.nineSlicedMode === SPRITE_RENDERMODE_TILED) {
            result += chunks.baseNineSlicedTiledPS;
        }
        return result;
    }

    // Add "Start" Code section to fragment shader.
    _fsGetStartCode(code, device, chunks, options) {
        let result = chunks.startPS;
        if (options.nineSlicedMode === SPRITE_RENDERMODE_SLICED) {
            result += chunks.startNineSlicedPS;
        } else if (options.nineSlicedMode === SPRITE_RENDERMODE_TILED) {
            result += chunks.startNineSlicedTiledPS;
        }
        return result;
    }

    // handles directional map shadow coordinate generation, including cascaded shadows
    _directionalShadowMapProjection(light, shadowCoordArgs, shadowParamArg, lightIndex, coordsFunctionName) {

        // for shadow cascades
        let code = "";
        if (light.numCascades > 1) {
            // compute which cascade matrix needs to be used
            code += `getShadowCascadeMatrix(light${lightIndex}_shadowMatrixPalette, light${lightIndex}_shadowCascadeDistances, light${lightIndex}_shadowCascadeCount);\n`;
            shadowCoordArgs = `(cascadeShadowMat, ${shadowParamArg});\n`;
        }

        // shadow coordinate generation
        code += coordsFunctionName + shadowCoordArgs;

        // stop shadow at the far distance
        code += `fadeShadow(light${lightIndex}_shadowCascadeDistances);\n`;
        return code;
    }

    _nonPointShadowMapProjection(device, light, shadowMatArg, shadowParamArg, lightIndex) {
        const shadowCoordArgs = `(${shadowMatArg}, ${shadowParamArg});\n`;
        if (!light._normalOffsetBias || light._isVsm) {
            if (light._type === LIGHTTYPE_SPOT) {
                if (light._isPcf && (device.webgl2 || device.extStandardDerivatives)) {
                    return "       getShadowCoordPerspZbuffer" + shadowCoordArgs;
                }
                return "       getShadowCoordPersp" + shadowCoordArgs;
            }
            return this._directionalShadowMapProjection(light, shadowCoordArgs, shadowParamArg, lightIndex, "getShadowCoordOrtho");
        }
        if (light._type === LIGHTTYPE_SPOT) {
            if (light._isPcf && (device.webgl2 || device.extStandardDerivatives)) {
                return "       getShadowCoordPerspZbufferNormalOffset" + shadowCoordArgs;
            }
            return "       getShadowCoordPerspNormalOffset" + shadowCoordArgs;
        }
        return this._directionalShadowMapProjection(light, shadowCoordArgs, shadowParamArg, lightIndex, "getShadowCoordOrthoNormalOffset");
    }

    _getLightSourceShapeString(shape) {
        switch (shape) {
            case LIGHTSHAPE_RECT:
                return 'Rect';
            case LIGHTSHAPE_DISK:
                return 'Disk';
            case LIGHTSHAPE_SPHERE:
                return 'Sphere';
            default:
                return '';
        }
    }

    generateVertexShader(useUv, useUnmodifiedUv, mapTransforms) {
        const device = this.device;
        const options = this.options;
        const chunks = this.chunks;

        let code = '';
        let codeBody = '';

        code += this._getPassDefineString(options.pass);

        // code += chunks.baseVS;
        code = this._vsAddBaseCode(code, chunks, options);

        codeBody += "   vPositionW    = getWorldPosition();\n";

        if (this.options.pass === SHADER_DEPTH) {
            code += 'varying float vDepth;\n';
            code += '#ifndef VIEWMATRIX\n';
            code += '#define VIEWMATRIX\n';
            code += 'uniform mat4 matrix_view;\n';
            code += '#endif\n';
            code += '#ifndef CAMERAPLANES\n';
            code += '#define CAMERAPLANES\n';
            code += 'uniform vec4 camera_params;\n\n';
            code += '#endif\n';
            codeBody += "    vDepth = -(matrix_view * vec4(vPositionW,1.0)).z * camera_params.x;\n";
        }

        if (this.options.useInstancing) {
            this.attributes.instance_line1 = SEMANTIC_ATTR12;
            this.attributes.instance_line2 = SEMANTIC_ATTR13;
            this.attributes.instance_line3 = SEMANTIC_ATTR14;
            this.attributes.instance_line4 = SEMANTIC_ATTR15;
            code += chunks.instancingVS;
        }

        if (this.needsNormal) {
            this.attributes.vertex_normal = SEMANTIC_NORMAL;
            codeBody += "   vNormalW = getNormal();\n";

            if (options.reflectionSource === 'sphereMap' && device.fragmentUniformsCount <= 16) {
                code += chunks.viewNormalVS;
                codeBody += "   vNormalV    = getViewNormal();\n";
            }

            if (options.hasTangents && (options.heightMap || options.normalMap || options.enableGGXSpecular)) {
                this.attributes.vertex_tangent = SEMANTIC_TANGENT;
                code += chunks.tangentBinormalVS;
                codeBody += "   vTangentW   = getTangent();\n";
                codeBody += "   vBinormalW  = getBinormal();\n";
            } else if (options.enableGGXSpecular) {
                code += chunks.tangentBinormalVS;
                codeBody += "   vObjectSpaceUpW  = getObjectSpaceUp();\n";
            }
        }

        const maxUvSets = 2;

        for (let i = 0; i < maxUvSets; i++) {
            if (useUv[i]) {
                this.attributes["vertex_texCoord" + i] = "TEXCOORD" + i;
                code += chunks["uv" + i + "VS"];
                codeBody += "   vec2 uv" + i + " = getUv" + i + "();\n";
            }
            if (useUnmodifiedUv[i]) {
                codeBody += "   vUv" + i + " = uv" + i + ";\n";
            }
        }

        /** @type {[string, string, string, any[]]} */
        const codes = [code, this.varyings, codeBody, []];

        mapTransforms.forEach((mapTransform) => {
            this._setMapTransform(codes, mapTransform.name, mapTransform.id, mapTransform.uv);
        });

        code = codes[0];
        this.varyings = codes[1];
        codeBody = codes[2];

        if (options.vertexColors) {
            this.attributes.vertex_color = SEMANTIC_COLOR;
            codeBody += "   vVertexColor = vertex_color;\n";
        }

        if (options.msdf && options.msdfTextAttribute) {
            this.attributes.vertex_outlineParameters = SEMANTIC_ATTR8;
            this.attributes.vertex_shadowParameters = SEMANTIC_ATTR9;

            codeBody += "    unpackMsdfParams();\n";

            code += chunks.msdfVS;
        }

        // morphing
        if (options.useMorphPosition || options.useMorphNormal) {

            if (options.useMorphTextureBased) {

                code += "#define MORPHING_TEXTURE_BASED\n";

                if (options.useMorphPosition) {
                    code += "#define MORPHING_TEXTURE_BASED_POSITION\n";
                }

                if (options.useMorphNormal) {
                    code += "#define MORPHING_TEXTURE_BASED_NORMAL\n";
                }

                // vertex ids attributes
                this.attributes.morph_vertex_id = SEMANTIC_ATTR15;
                code += "attribute float morph_vertex_id;\n";

            } else {

                // set up 8 slots for morphing. these are supported combinations: PPPPPPPP, NNNNNNNN, PPPPNNNN
                code += "#define MORPHING\n";

                // first 4 slots are either position or normal
                if (options.useMorphPosition) {
                    this.attributes.morph_pos0 = SEMANTIC_ATTR8;
                    this.attributes.morph_pos1 = SEMANTIC_ATTR9;
                    this.attributes.morph_pos2 = SEMANTIC_ATTR10;
                    this.attributes.morph_pos3 = SEMANTIC_ATTR11;

                    code += "#define MORPHING_POS03\n";
                    code += "attribute vec3 morph_pos0;\n";
                    code += "attribute vec3 morph_pos1;\n";
                    code += "attribute vec3 morph_pos2;\n";
                    code += "attribute vec3 morph_pos3;\n";

                } else if (options.useMorphNormal) {
                    this.attributes.morph_nrm0 = SEMANTIC_ATTR8;
                    this.attributes.morph_nrm1 = SEMANTIC_ATTR9;
                    this.attributes.morph_nrm2 = SEMANTIC_ATTR10;
                    this.attributes.morph_nrm3 = SEMANTIC_ATTR11;

                    code += "#define MORPHING_NRM03\n";
                    code += "attribute vec3 morph_nrm0;\n";
                    code += "attribute vec3 morph_nrm1;\n";
                    code += "attribute vec3 morph_nrm2;\n";
                    code += "attribute vec3 morph_nrm3;\n";
                }

                // next 4 slots are either position or normal
                if (!options.useMorphNormal) {
                    this.attributes.morph_pos4 = SEMANTIC_ATTR12;
                    this.attributes.morph_pos5 = SEMANTIC_ATTR13;
                    this.attributes.morph_pos6 = SEMANTIC_ATTR14;
                    this.attributes.morph_pos7 = SEMANTIC_ATTR15;

                    code += "#define MORPHING_POS47\n";
                    code += "attribute vec3 morph_pos4;\n";
                    code += "attribute vec3 morph_pos5;\n";
                    code += "attribute vec3 morph_pos6;\n";
                    code += "attribute vec3 morph_pos7;\n";
                } else {
                    this.attributes.morph_nrm4 = SEMANTIC_ATTR12;
                    this.attributes.morph_nrm5 = SEMANTIC_ATTR13;
                    this.attributes.morph_nrm6 = SEMANTIC_ATTR14;
                    this.attributes.morph_nrm7 = SEMANTIC_ATTR15;

                    code += "#define MORPHING_NRM47\n";
                    code += "attribute vec3 morph_nrm4;\n";
                    code += "attribute vec3 morph_nrm5;\n";
                    code += "attribute vec3 morph_nrm6;\n";
                    code += "attribute vec3 morph_nrm7;\n";
                }
            }
        }

        if (options.skin) {
            this.attributes.vertex_boneWeights = SEMANTIC_BLENDWEIGHT;
            this.attributes.vertex_boneIndices = SEMANTIC_BLENDINDICES;
            code += skinCode(device, chunks);
            code += "#define SKIN\n";
        } else if (options.useInstancing) {
            code += "#define INSTANCING\n";
        }
        if (options.screenSpace) {
            code += "#define SCREENSPACE\n";
        }
        if (options.pixelSnap) {
            code += "#define PIXELSNAP\n";
        }

        code = this._vsAddTransformCode(code, device, chunks, options);

        if (this.needsNormal) {
            code += chunks.normalVS;
        }

        code += "\n";
        code += chunks.startVS;
        code += codeBody;
        code += chunks.endVS;
        code += "}";

        // build varyings
        Object.keys(builtinVaryings).forEach((v) => {
            if (code.indexOf(v) >= 0) {
                this.varyings += `varying ${builtinVaryings[v]} ${v};\n`;
            }
        });

        let startCode = "";
        if (device.webgl2) {
            startCode = versionCode(device);
            if (chunks.extensionVS) {
                startCode += chunks.extensionVS + "\n";
            }
            startCode += chunks.gles3VS;
        } else {
            if (chunks.extensionVS) {
                startCode = chunks.extensionVS + "\n";
            }
        }

        this.vshader = startCode + this.varyings + code;
    }

    _fsGetBeginCode() {
        let code = '';

        if (this.device.webgl2) {
            code += versionCode(this.device);
        } else {
            if (this.device.extStandardDerivatives) {
                code += "#extension GL_OES_standard_derivatives : enable\n";
            }
            if (this.device.extTextureLod) {
                code += "#extension GL_EXT_shader_texture_lod : enable\n";
                code += "#define SUPPORTS_TEXLOD\n";
            }
        }

        if (this.chunks.extensionPS) {
            code += this.chunks.extensionPS + "\n";
        }

        if (this.device.webgl2) {
            code += this.chunks.gles3PS;
        }

        code += this.options.forceFragmentPrecision ? "precision " + this.options.forceFragmentPrecision + " float;\n\n" : precisionCode(this.device);
        code += this._getPassDefineString(this.options.pass);

        return code;
    }

    _fsGetPickPassCode() {
        let code = this._fsGetBeginCode();
        code += "uniform vec4 uColor;\n";
        code += this.varyings;
        code += this.frontendDecl;
        code += this.frontendCode;
        code += begin();
        code += this.frontendFunc;
        code += "    gl_FragColor = uColor;\n";
        code += end();
        return code;
    }

    _fsGetDepthPassCode() {
        const chunks = this.chunks;

        let code = this._fsGetBeginCode();

        code += 'varying float vDepth;\n';
        code += this.varyings;
        code += chunks.packDepthPS;
        code += this.frontendDecl;
        code += this.frontendCode;
        code += begin();
        code += this.frontendFunc;
        code += "    gl_FragColor = packFloat(vDepth);\n";
        code += end();

        return code;
    }

    _fsGetShadowPassCode() {
        const device = this.device;
        const options = this.options;
        const chunks = this.chunks;
        const varyings = this.varyings;

        const lightType = ShaderPass.toLightType(options.pass);
        const shadowType = ShaderPass.toShadowType(options.pass);

        let code = this._fsGetBeginCode();

        if (device.extStandardDerivatives && !device.webgl2) {
            code += 'uniform vec2 polygonOffset;\n';
        }

        if (shadowType === SHADOW_VSM32) {
            if (device.textureFloatHighPrecision) {
                code += '#define VSM_EXPONENT 15.0\n\n';
            } else {
                code += '#define VSM_EXPONENT 5.54\n\n';
            }
        } else if (shadowType === SHADOW_VSM16) {
            code += '#define VSM_EXPONENT 5.54\n\n';
        }

        if (lightType !== LIGHTTYPE_DIRECTIONAL) {
            code += 'uniform vec3 view_position;\n';
            code += 'uniform float light_radius;\n';
        }

        code += varyings;
        code += this.frontendDecl;
        code += this.frontendCode;

        if (shadowType === SHADOW_PCF3 && (!device.webgl2 || lightType === LIGHTTYPE_OMNI)) {
            code += chunks.packDepthPS;
        } else if (shadowType === SHADOW_VSM8) {
            code += "vec2 encodeFloatRG( float v ) {\n";
            code += "    vec2 enc = vec2(1.0, 255.0) * v;\n";
            code += "    enc = fract(enc);\n";
            code += "    enc -= enc.yy * vec2(1.0/255.0, 1.0/255.0);\n";
            code += "    return enc;\n";
            code += "}\n\n";
        }

        code += begin();

        code += this.frontendFunc;

        const isVsm = shadowType === SHADOW_VSM8 || shadowType === SHADOW_VSM16 || shadowType === SHADOW_VSM32;

        if (lightType === LIGHTTYPE_OMNI || (isVsm && lightType !== LIGHTTYPE_DIRECTIONAL)) {
            code += "    float depth = min(distance(view_position, vPositionW) / light_radius, 0.99999);\n";
        } else {
            code += "    float depth = gl_FragCoord.z;\n";
        }

        if (shadowType === SHADOW_PCF3 && (!device.webgl2 || (lightType === LIGHTTYPE_OMNI && !options.clusteredLightingEnabled))) {
            if (device.extStandardDerivatives && !device.webgl2) {
                code += "    float minValue = 2.3374370500153186e-10; //(1.0 / 255.0) / (256.0 * 256.0 * 256.0);\n";
                code += "    depth += polygonOffset.x * max(abs(dFdx(depth)), abs(dFdy(depth))) + minValue * polygonOffset.y;\n";
                code += "    gl_FragColor = packFloat(depth);\n";
            } else {
                code += "    gl_FragColor = packFloat(depth);\n";
            }
        } else if (shadowType === SHADOW_PCF3 || shadowType === SHADOW_PCF5) {
            code += "    gl_FragColor = vec4(1.0);\n"; // just the simplest code, color is not written anyway

            // clustered omni light is using shadow sampler and needs to write custom depth
            if (options.clusteredLightingEnabled && lightType === LIGHTTYPE_OMNI && device.webgl2) {
                code += "    gl_FragDepth = depth;\n";
            }
        } else if (shadowType === SHADOW_VSM8) {
            code += "    gl_FragColor = vec4(encodeFloatRG(depth), encodeFloatRG(depth*depth));\n";
        } else {
            code += chunks.storeEVSMPS;
        }

        code += end();

        return code;
    }

    // the big one
    _fsGetLitPassCode() {
        const device = this.device;
        const options = this.options;
        const chunks = this.chunks;

        let code = "";

        if (options.clearCoat > 0) {
            code += '#define CLEARCOAT\n';

            // enable clear-coat path in clustered chunk
            code += "#define CLUSTER_CLEAR_COAT\n";
        }

        if (options.opacityFadesSpecular === false) {
            code += 'uniform float material_alphaFade;\n';
        }

        // FRAGMENT SHADER INPUTS: UNIFORMS
        const shadowTypeUsed = [];
        let numShadowLights = 0;
        let shadowedDirectionalLightUsed = false;
        let useVsm = false;
        let usePerspZbufferShadow = false;

        let hasAreaLights = options.lights.some(function (light) {
            return light._shape && light._shape !== LIGHTSHAPE_PUNCTUAL;
        });

        // if clustered lighting has area lights enabled, it always runs in 'area lights mode'
        // TODO: maybe we should always use it and remove the other way?
        if (options.clusteredLightingEnabled && options.clusteredLightingAreaLightsEnabled) {
            hasAreaLights = true;
        }

        if (device.areaLightLutFormat === PIXELFORMAT_R8_G8_B8_A8) {
            // use offset and scale for rgb8 format luts
            code += "#define AREA_R8_G8_B8_A8_LUTS\n";
            code += "#define AREA_LUTS_PRECISION lowp\n";
        } else {
            code += "#define AREA_LUTS_PRECISION highp\n";
        }

        if (hasAreaLights || options.clusteredLightingEnabled) {
            code += "#define AREA_LIGHTS\n";
            code += "uniform AREA_LUTS_PRECISION sampler2D areaLightsLutTex1;\n";
            code += "uniform AREA_LUTS_PRECISION sampler2D areaLightsLutTex2;\n";
        }

        for (let i = 0; i < options.lights.length; i++) {
            const light = options.lights[i];
            const lightType = light._type;

            // skip uniform generation for local lights if clustered lighting is enabled
            if (options.clusteredLightingEnabled && lightType !== LIGHTTYPE_DIRECTIONAL)
                continue;

            const lightShape = (hasAreaLights && light._shape) ? light._shape : LIGHTSHAPE_PUNCTUAL;

            code += "uniform vec3 light" + i + "_color;\n";
            if (lightType === LIGHTTYPE_DIRECTIONAL) {
                code += "uniform vec3 light" + i + "_direction;\n";
            } else {
                code += "uniform vec3 light" + i + "_position;\n";
                code += "uniform float light" + i + "_radius;\n";
                if (lightType === LIGHTTYPE_SPOT) {
                    code += "uniform vec3 light" + i + "_direction;\n";
                    code += "uniform float light" + i + "_innerConeAngle;\n";
                    code += "uniform float light" + i + "_outerConeAngle;\n";
                }
            }
            if (lightShape !== LIGHTSHAPE_PUNCTUAL) {
                if (lightType === LIGHTTYPE_DIRECTIONAL) {
                    code += "uniform vec3 light" + i + "_position;\n";
                }
                code += "uniform vec3 light" + i + "_halfWidth;\n";
                code += "uniform vec3 light" + i + "_halfHeight;\n";
            }
            if (light.castShadows && !options.noShadow) {
                code += "uniform mat4 light" + i + "_shadowMatrix;\n";
                code += "uniform float light" + i + "_shadowIntensity;\n";

                // directional (cascaded) shadows
                if (lightType === LIGHTTYPE_DIRECTIONAL) {
                    code += "uniform mat4 light" + i + "_shadowMatrixPalette[4];\n";
                    code += "uniform float light" + i + "_shadowCascadeDistances[4];\n";
                    code += "uniform float light" + i + "_shadowCascadeCount;\n";
                }

                if (lightType !== LIGHTTYPE_DIRECTIONAL) {
                    code += "uniform vec4 light" + i + "_shadowParams;\n"; // Width, height, bias, radius
                } else {
                    shadowedDirectionalLightUsed = true;
                    code += "uniform vec3 light" + i + "_shadowParams;\n"; // Width, height, bias
                }
                if (lightType === LIGHTTYPE_OMNI) {
                    code += "uniform samplerCube light" + i + "_shadowMap;\n";
                } else {
                    if (light._isPcf && device.webgl2) {
                        code += "uniform sampler2DShadow light" + i + "_shadowMap;\n";
                    } else {
                        code += "uniform sampler2D light" + i + "_shadowMap;\n";
                    }
                }
                numShadowLights++;
                shadowTypeUsed[light._shadowType] = true;
                if (light._isVsm) useVsm = true;
                if (light._isPcf && (device.webgl2 || device.extStandardDerivatives) && lightType === LIGHTTYPE_SPOT) usePerspZbufferShadow = true;
            }
            if (light._cookie) {
                if (light._cookie._cubemap) {
                    if (lightType === LIGHTTYPE_OMNI) {
                        code += "uniform samplerCube light" + i + "_cookie;\n";
                        code += "uniform float light" + i + "_cookieIntensity;\n";
                        if (!light.castShadows || options.noShadow) code += "uniform mat4 light" + i + "_shadowMatrix;\n";
                    }
                } else {
                    if (lightType === LIGHTTYPE_SPOT) {
                        code += "uniform sampler2D light" + i + "_cookie;\n";
                        code += "uniform float light" + i + "_cookieIntensity;\n";
                        if (!light.castShadows || options.noShadow) code += "uniform mat4 light" + i + "_shadowMatrix;\n";
                        if (light._cookieTransform) {
                            code += "uniform vec4 light" + i + "_cookieMatrix;\n";
                            code += "uniform vec2 light" + i + "_cookieOffset;\n";
                        }
                    }
                }
            }
        }

        code += "\n"; // End of uniform declarations

        // TBN
        const hasTBN = this.needsNormal && (options.normalMap || options.clearCoatNormalMap || (options.enableGGXSpecular && !options.heightMap));

        if (hasTBN) {
            if (options.hasTangents) {
                code += options.fastTbn ? chunks.TBNfastPS : chunks.TBNPS;
            } else {
                if (device.extStandardDerivatives && (options.normalMap || options.clearCoatNormalMap)) {
                    code += chunks.TBNderivativePS.replace(/\$UV/g, this.lightingUv);
                } else {
                    code += chunks.TBNObjectSpacePS;
                }
            }
        }

        // FIXME: only add these when needed
        code += chunks.sphericalPS;
        code += chunks.decodePS;
        code += gammaCode(options.gamma, chunks);
        code += tonemapCode(options.toneMap, chunks);
        code += fogCode(options.fog, chunks);


        // frontend
        code += this.frontendCode;

        if (options.useCubeMapRotation) {
            code += "#define CUBEMAP_ROTATION\n";
        }

        if (this.needsNormal) {
            code += chunks.cubeMapRotatePS;
            code += options.cubeMapProjection > 0 ? chunks.cubeMapProjectBoxPS : chunks.cubeMapProjectNonePS;
            code += options.skyboxIntensity ? chunks.envMultiplyPS : chunks.envConstPS;
        }

        if ((this.lighting && options.useSpecular) || this.reflections) {
            if (options.useMetalness) {
                code += chunks.metalnessModulatePS;
            }

            if (options.fresnelModel === FRESNEL_SCHLICK) {
                code += chunks.fresnelSchlickPS;
            }
        }

        const useAo = options.aoMap || options.aoVertexColor;

        if (useAo) {
            code += chunks.aoDiffuseOccPS;
            switch (options.occludeSpecular) {
                case SPECOCC_AO:
                    code += options.occludeSpecularFloat ? chunks.aoSpecOccSimplePS : chunks.aoSpecOccConstSimplePS;
                    break;
                case SPECOCC_GLOSSDEPENDENT:
                    code += options.occludeSpecularFloat ? chunks.aoSpecOccPS : chunks.aoSpecOccConstPS;
                    break;
                default:
                    break;
            }
        }

        if (options.reflectionSource === 'envAtlas') {
            code += chunks.envAtlasPS;
            code += chunks.reflectionEnvPS.replace(/\$DECODE/g, ChunkUtils.decodeFunc(options.reflectionEncoding));
        } else if (options.reflectionSource === 'cubeMap') {
            code += options.fixSeams ? chunks.fixCubemapSeamsStretchPS : chunks.fixCubemapSeamsNonePS;
            code += chunks.reflectionCubePS.replace(/\$DECODE/g, ChunkUtils.decodeFunc(options.reflectionEncoding));
        } else if (options.reflectionSource === 'sphereMap') {
            const scode = device.fragmentUniformsCount > 16 ? chunks.reflectionSpherePS : chunks.reflectionSphereLowPS;
            code += scode.replace(/\$DECODE/g, ChunkUtils.decodeFunc(options.reflectionEncoding));
        }

        if (this.reflections) {
            if (options.clearCoat > 0) {
                code += chunks.reflectionCCPS;
            }
            if (options.refraction) {
                code += chunks.refractionPS;
            }
        }

        // clustered lighting
        if (options.clusteredLightingEnabled) {

            // include this before shadow / cookie code
            code += chunks.clusteredLightUtilsPS;
            code += chunks.clusteredLightCookiesPS;

            // always include shadow chunks clustered lights support
            shadowTypeUsed[SHADOW_PCF3] = true;
            shadowTypeUsed[SHADOW_PCF5] = true;
            usePerspZbufferShadow = true;
        }

        if (numShadowLights > 0 || options.clusteredLightingEnabled) {
            if (shadowedDirectionalLightUsed) {
                code += chunks.shadowCascadesPS;
            }
            if (shadowTypeUsed[SHADOW_PCF3]) {
                code += chunks.shadowStandardPS;
            }
            if (shadowTypeUsed[SHADOW_PCF5] && device.webgl2) {
                code += chunks.shadowStandardGL2PS;
            }
            if (useVsm) {
                code += chunks.shadowVSM_commonPS;
                if (shadowTypeUsed[SHADOW_VSM8]) {
                    code += chunks.shadowVSM8PS;
                }
                if (shadowTypeUsed[SHADOW_VSM16]) {
                    code += device.extTextureHalfFloatLinear ? chunks.shadowEVSMPS.replace(/\$/g, "16") : chunks.shadowEVSMnPS.replace(/\$/g, "16");
                }
                if (shadowTypeUsed[SHADOW_VSM32]) {
                    code += device.extTextureFloatLinear ? chunks.shadowEVSMPS.replace(/\$/g, "32") : chunks.shadowEVSMnPS.replace(/\$/g, "32");
                }
            }

            if (!(device.webgl2 || device.extStandardDerivatives)) {
                code += chunks.biasConstPS;
            }

            // otherwise bias is applied on render
            code += chunks.shadowCoordPS + chunks.shadowCommonPS;
            if (usePerspZbufferShadow) code += chunks.shadowCoordPerspZbufferPS;
        }

        if (options.enableGGXSpecular) code += "uniform float material_anisotropy;\n";

        if (this.lighting) {
            code += chunks.lightDiffuseLambertPS;
            if (hasAreaLights || options.clusteredLightingEnabled) code += chunks.ltc;
        }

        code += '\n';

        let useOldAmbient = false;
        if (options.useSpecular) {

            // enable specular path in clustered chunk
            code += "#define CLUSTER_SPECULAR\n";

            if (options.fresnelModel > 0) {
                code += '#define CLUSTER_SPECULAR_FRESNEL\n';
            }

            // enable conserve energy path in clustered chunk
            if (options.conserveEnergy) {
                code += "#define CLUSTER_CONSERVE_ENERGY\n";
            }

            if (this.lighting) {
                code += options.shadingModel === SPECULAR_PHONG ? chunks.lightSpecularPhongPS : (options.enableGGXSpecular ? chunks.lightSpecularAnisoGGXPS : chunks.lightSpecularBlinnPS);
            }

            if (options.fresnelModel > 0) {
                // schlick
                code += chunks.combineDiffuseSpecularPS;
            } else {
                // no fresnel
                if (this.reflections) {
                    code += chunks.combineDiffuseSpecularOldPS;
                } else {
                    // no reflections
                    // FIXME: the following test for presence of diffuseMap and then resulting
                    // chunk selection seems arbitrary; why would the presence of diffuseMap
                    // indicate "useOldAmbient"?
                    if (options.diffuseMap) {
                        code += chunks.combineDiffuseSpecularNoReflPS; // if you don't use environment cubemaps, you may consider this
                    } else {
                        code += chunks.combineDiffuseSpecularNoReflSeparateAmbientPS;
                        useOldAmbient = true;
                    }
                }
            }
        } else {
            code += chunks.combineDiffusePS;
        }

        if (options.clearCoat > 0) {
            code += chunks.combineClearCoatPS;
        }

        // lightmap support
        if (options.lightMap || options.lightVertexColor) {
            code += (options.useSpecular && options.dirLightMap) ? chunks.lightmapDirAddPS : chunks.lightmapAddPS;
        }

        const addAmbient = (!options.lightMap && !options.lightVertexColor) || options.lightMapWithoutAmbient;

        if (addAmbient) {
            if (options.ambientSource === 'ambientSH') {
                code += chunks.ambientSHPS;
            } else if (options.ambientSource === 'envAtlas') {
                if (options.reflectionSource !== 'envAtlas') {
                    code += chunks.envAtlasPS;
                }
                code += chunks.ambientEnvPS.replace(/\$DECODE/g, ChunkUtils.decodeFunc(options.ambientEncoding));
            } else {
                code += chunks.ambientConstantPS;
            }
        }

        if (options.ambientTint && !useOldAmbient) {
            code += "uniform vec3 material_ambient;\n";
        }

        if (options.msdf) {
            if (!options.msdfTextAttribute) {
                code += "\n#define UNIFORM_TEXT_PARAMETERS";
            }
            code += chunks.msdfPS;
        }

        if (this.needsNormal) {
            code += chunks.viewDirPS;
            if (options.useSpecular) {
                code += (options.enableGGXSpecular) ? chunks.reflDirAnisoPS : chunks.reflDirPS;
            }
        }

        let hasPointLights = false;
        let usesLinearFalloff = false;
        let usesInvSquaredFalloff = false;
        let usesSpot = false;
        let usesCookie = false;
        let usesCookieNow;

        // clustered lighting
        if (options.clusteredLightingEnabled && this.lighting) {

            usesSpot = true;
            hasPointLights = true;
            usesLinearFalloff = true;
            usesCookie = true;

            code += chunks.floatUnpackingPS;

            if (options.lightMaskDynamic)
                code += "\n#define CLUSTER_MESH_DYNAMIC_LIGHTS";

            if (options.clusteredLightingCookiesEnabled)
                code += "\n#define CLUSTER_COOKIES";
            if (options.clusteredLightingShadowsEnabled && !options.noShadow) {
                code += "\n#define CLUSTER_SHADOWS";
                code += "\n#define CLUSTER_SHADOW_TYPE_" + shadowTypeToString[options.clusteredLightingShadowType];
            }

            if (options.clusteredLightingAreaLightsEnabled)
                code += "\n#define CLUSTER_AREALIGHTS";

            code += LightsBuffer.shaderDefines;
            code += chunks.clusteredLightShadowsPS;
            code += chunks.clusteredLightPS;
        }

        if (options.twoSidedLighting) code += "uniform float twoSidedLightingNegScaleFactor;\n";

        // FRAGMENT SHADER BODY

        code += this._fsGetStartCode(code, device, chunks, options);

        if (this.needsNormal) {
            if (options.twoSidedLighting) {
                code += "    dVertexNormalW = normalize(gl_FrontFacing ? vNormalW * twoSidedLightingNegScaleFactor : -vNormalW * twoSidedLightingNegScaleFactor);\n";
            } else {
                code += "    dVertexNormalW = normalize(vNormalW);\n";
            }

            if ((options.heightMap || options.normalMap) && options.hasTangents) {
                if (options.twoSidedLighting) {
                    code += "    dTangentW = gl_FrontFacing ? vTangentW * twoSidedLightingNegScaleFactor : -vTangentW * twoSidedLightingNegScaleFactor;\n";
                    code += "    dBinormalW = gl_FrontFacing ? vBinormalW * twoSidedLightingNegScaleFactor : -vBinormalW * twoSidedLightingNegScaleFactor;\n";
                } else {
                    code += "    dTangentW = vTangentW;\n";
                    code += "    dBinormalW = vBinormalW;\n";
                }
            }

            code += "    getViewDir();\n";
            if (hasTBN) {
                code += "    getTBN();\n";
            }
        }

        // invoke frontend functions
        code += this.frontendFunc;

        // transform tangent space normals to world space
        if (this.needsNormal) {
            if (options.useSpecular) {
                code += "    getReflDir();\n";
            }

            if (options.clearCoat > 0) {
                code += "    ccReflDirW = normalize(-reflect(dViewDirW, ccNormalW));\n";
            }
        }

        if ((this.lighting && options.useSpecular) || this.reflections) {

            if (options.useMetalness) {
                code += "    getMetalnessModulate();\n";
            }
        }

        if (addAmbient) {
            code += "    addAmbient();\n";

            // move ambient color out of diffuse (used by Lightmapper, to multiply ambient color by accumulated AO)
            if (options.separateAmbient) {
                code += `
                    vec3 dAmbientLight = dDiffuseLight;
                    dDiffuseLight = vec3(0);
                `;
            }
        }

        if (options.ambientTint && !useOldAmbient) {
            code += "    dDiffuseLight *= material_ambient;\n";
        }

        if (useAo && !options.occludeDirect) {
            code += "    occludeDiffuse();\n";
        }

        if (options.lightMap || options.lightVertexColor) {
            code += "    addLightMap();\n";
        }

        if (this.lighting || this.reflections) {
            if (this.reflections) {
                if (options.clearCoat > 0) {
                    code += "    addReflectionCC();\n";
                    if (options.fresnelModel > 0) {
                        code += "    ccReflection.rgb *= getFresnel(dot(dViewDirW, ccNormalW), vec3(ccSpecularity));\n";
                    }  else {
                        code += "    ccReflection.rgb *= ccSpecularity;\n";
                    }
                }
                if (options.useSpecularityFactor) {
                    code += "    ccReflection.rgb *= dSpecularityFactor;\n";
                }
                code += "    addReflection();\n";

                // Fresnel has to be applied to reflections
                if (options.fresnelModel > 0) {
                    code += "    dReflection.rgb *= getFresnel(dot(dViewDirW, dNormalW), dSpecularity);\n";
                } else {
                    code += "    dReflection.rgb *= dSpecularity;\n";
                }
                if (options.useSpecularityFactor) {
                    code += "    dReflection.rgb *= dSpecularityFactor;\n";
                }
            }

            if (hasAreaLights) {
                // specular has to be accumulated differently if we want area lights to look correct
                code += "    dSpecularLight *= dSpecularity;\n";
                // code += "    float roughness = max((1.0 - dGlossiness) * (1.0 - dGlossiness), 0.001);\n";

                // evaluate material based area lights data, shared by all area lights
                if (options.useSpecular) {
                    code += "    calcLTCLightValues();\n";
                }
            }

            for (let i = 0; i < options.lights.length; i++) {
                const light = options.lights[i];
                const lightType = light._type;

                // if clustered lights are used, skip normal lights other than directional
                if (options.clusteredLightingEnabled && lightType !== LIGHTTYPE_DIRECTIONAL) {
                    continue;
                }

                // The following code is not decoupled to separate shader files, because most of it can be actually changed to achieve different behaviors like:
                // - different falloffs
                // - different shadow coords (omni shadows will use drastically different genShadowCoord)
                // - different shadow filter modes
                // - different light source shapes

                // getLightDiffuse and getLightSpecular is BRDF itself.

                usesCookieNow = false;

                const lightShape = (hasAreaLights && light._shape) ? light.shape : LIGHTSHAPE_PUNCTUAL;
                const shapeString = (hasAreaLights && light._shape) ? this._getLightSourceShapeString(lightShape) : '';

                if (lightShape !== LIGHTSHAPE_PUNCTUAL) {
                    code += "    calc" + shapeString + "LightValues(light" + i + "_position, light" + i + "_halfWidth, light" + i + "_halfHeight);\n";
                }

                if (lightType === LIGHTTYPE_DIRECTIONAL) {
                    // directional
                    code += "    dLightDirNormW = light" + i + "_direction;\n";
                    code += "    dAtten = 1.0;\n";
                } else {

                    if (light._cookie) {
                        if (lightType === LIGHTTYPE_SPOT && !light._cookie._cubemap) {
                            usesCookie = true;
                            usesCookieNow = true;
                        } else if (lightType === LIGHTTYPE_OMNI && light._cookie._cubemap) {
                            usesCookie = true;
                            usesCookieNow = true;
                        }
                    }

                    code += "    getLightDirPoint(light" + i + "_position);\n";
                    hasPointLights = true;

                    if (usesCookieNow) {
                        if (lightType === LIGHTTYPE_SPOT) {
                            code += "    dAtten3 = getCookie2D" + (light._cookieFalloff ? "" : "Clip") + (light._cookieTransform ? "Xform" : "") + "(light" + i + "_cookie, light" + i + "_shadowMatrix, light" + i + "_cookieIntensity" + (light._cookieTransform ? ", light" + i + "_cookieMatrix, light" + i + "_cookieOffset" : "") + ")." + light._cookieChannel + ";\n";
                        } else {
                            code += "    dAtten3 = getCookieCube(light" + i + "_cookie, light" + i + "_shadowMatrix, light" + i + "_cookieIntensity)." + light._cookieChannel + ";\n";
                        }
                    }

                    if (lightShape === LIGHTSHAPE_PUNCTUAL) {
                        if (light._falloffMode === LIGHTFALLOFF_LINEAR) {
                            code += "    dAtten = getFalloffLinear(light" + i + "_radius);\n";
                            usesLinearFalloff = true;
                        } else {
                            code += "    dAtten = getFalloffInvSquared(light" + i + "_radius);\n";
                            usesInvSquaredFalloff = true;
                        }
                    } else {
                        // non punctual lights only gets the range window here
                        code += "    dAtten = getFalloffWindow(light" + i + "_radius);\n";
                        usesInvSquaredFalloff = true;
                    }

                    code += "    if (dAtten > 0.00001) {\n"; // BRANCH START

                    if (lightType === LIGHTTYPE_SPOT) {
                        if (!(usesCookieNow && !light._cookieFalloff)) {
                            code += "    dAtten *= getSpotEffect(light" + i + "_direction, light" + i + "_innerConeAngle, light" + i + "_outerConeAngle);\n";
                            usesSpot = true;
                        }
                    }
                }

                // diffuse lighting - LTC lights do not mix diffuse lighting into attenuation that affects specular
                if (lightShape !== LIGHTSHAPE_PUNCTUAL) {
                    if (lightType === LIGHTTYPE_DIRECTIONAL) {
                        // NB: A better aproximation perhaps using wrap lighting could be implemented here
                        code += "    dAttenD = getLightDiffuse();\n";
                    } else {
                        // 16.0 is a constant that is in getFalloffInvSquared()
                        code += "    dAttenD = get" + shapeString + "LightDiffuse() * 16.0;\n";
                    }
                } else {
                    code += "    dAtten *= getLightDiffuse();\n";
                }

                if (light.castShadows && !options.noShadow) {
                    let shadowReadMode = null;
                    let evsmExp;
                    if (light._shadowType === SHADOW_VSM8) {
                        shadowReadMode = "VSM8";
                        evsmExp = "0.0";
                    } else if (light._shadowType === SHADOW_VSM16) {
                        shadowReadMode = "VSM16";
                        evsmExp = "5.54";
                    } else if (light._shadowType === SHADOW_VSM32) {
                        shadowReadMode = "VSM32";
                        if (device.textureFloatHighPrecision) {
                            evsmExp = "15.0";
                        } else {
                            evsmExp = "5.54";
                        }
                    } else if (light._shadowType === SHADOW_PCF5) {
                        shadowReadMode = "PCF5x5";
                    } else {
                        shadowReadMode = "PCF3x3";
                    }

                    if (shadowReadMode !== null) {
                        if (lightType === LIGHTTYPE_OMNI) {
                            const shadowCoordArgs = "(light" + i + "_shadowMap, light" + i + "_shadowParams);\n";
                            if (light._normalOffsetBias) {
                                code += "    normalOffsetPointShadow(light" + i + "_shadowParams);\n";
                            }
                            code += `    float shadow${i} = getShadowPoint${shadowReadMode}${shadowCoordArgs}`;
                            code += `    dAtten *= mix(1.0, shadow${i}, light${i}_shadowIntensity);\n`;
                        } else {
                            const shadowMatArg = `light${i}_shadowMatrix`;
                            const shadowParamArg = `light${i}_shadowParams`;
                            code += this._nonPointShadowMapProjection(device, options.lights[i], shadowMatArg, shadowParamArg, i);

                            if (lightType === LIGHTTYPE_SPOT) shadowReadMode = "Spot" + shadowReadMode;
                            code += `    float shadow${i} = getShadow${shadowReadMode}(light${i}_shadowMap, light${i}_shadowParams${(light._isVsm ? ", " + evsmExp : "")});\n`;
                            code += `    dAtten *= mix(1.0, shadow${i}, light${i}_shadowIntensity);\n`;
                        }
                    }
                }

                if (lightShape !== LIGHTSHAPE_PUNCTUAL) {

                    // area light - they do not mix diffuse lighting into specular attenuation
                    if (options.conserveEnergy && options.useSpecular) {
                        code += "    dDiffuseLight += mix((dAttenD * dAtten) * light" + i + "_color" + (usesCookieNow ? " * dAtten3" : "") + ", vec3(0), dLTCSpecFres);\n";
                    } else {
                        code += "    dDiffuseLight += (dAttenD * dAtten) * light" + i + "_color" + (usesCookieNow ? " * dAtten3" : "") + ";\n";
                    }
                } else {

                    // punctual light
                    if (hasAreaLights && options.conserveEnergy && options.useSpecular) {
                        code += "    dDiffuseLight += mix(dAtten * light" + i + "_color" + (usesCookieNow ? " * dAtten3" : "") + ", vec3(0), dSpecularity);\n";
                    } else {
                        code += "    dDiffuseLight += dAtten * light" + i + "_color" + (usesCookieNow ? " * dAtten3" : "") + ";\n";
                    }
                }

                if (options.useSpecular) {
                    code += "    dHalfDirW = normalize(-dLightDirNormW + dViewDirW);\n";
                }

                // specular / clear coat
                if (lightShape !== LIGHTSHAPE_PUNCTUAL) {

                    // area light
                    if (options.clearCoat > 0) code += "    ccSpecularLight += ccLTCSpecFres * get" + shapeString + "LightSpecularCC() * dAtten * light" + i + "_color" + (usesCookieNow ? " * dAtten3" : "") + ";\n";
                    if (options.useSpecular) code += "    dSpecularLight += dLTCSpecFres * get" + shapeString + "LightSpecular() * dAtten * light" + i + "_color" + (usesCookieNow ? " * dAtten3" : "") + ";\n";

                } else {
                    var calcFresnel = false;
                    if (lightType === LIGHTTYPE_DIRECTIONAL && options.fresnelModel > 0) {
                        calcFresnel = true;
                    }

                    // if LTC lights are present, specular must be accumulated with specularity (specularity is pre multiplied by punctual light fresnel)
                    if (options.clearCoat > 0) {
                        code += "    ccSpecularLight += getLightSpecularCC(dHalfDirW) * dAtten * light" + i + "_color";
                        code += usesCookieNow ? " * dAtten3" : "";
                        code += calcFresnel ? " * getFresnel(dot(dViewDirW, dHalfDirW), vec3(ccSpecularity))" : " * vec3(ccSpecularity)";
                        code +=  ";\n";
                    }
                    if (options.useSpecular) {
                        code += "    dSpecularLight += getLightSpecular(dHalfDirW) * dAtten * light" + i + "_color";
                        code += usesCookieNow ? " * dAtten3" : "";
                        code += calcFresnel ? " * getFresnel(dot(dViewDirW, dHalfDirW), dSpecularity)" : " * dSpecularity";
                        code += ";\n";
                    }
                }

                if (lightType !== LIGHTTYPE_DIRECTIONAL) {
                    code += "    }\n"; // BRANCH END
                }

                code += "\n";
            }

            // clustered lighting
            if (options.clusteredLightingEnabled && this.lighting) {
                usesLinearFalloff = true;
                usesInvSquaredFalloff = true;
                hasPointLights = true;
                code += "    addClusteredLights();\n";
            }

            if (hasAreaLights) {
                // specular has to be accumulated differently if we want area lights to look correct
                if (options.clearCoat > 0) {
                    code += "    ccSpecularity = 1.0;\n";
                }
                if (options.useSpecular) {
                    code += "    dSpecularity = vec3(1);\n";
                }
            }

            if (this.reflections && options.refraction) {
                code += "    addRefraction();\n";
            }
        }
        code += "\n";

        if (useAo) {
            if (options.occludeDirect) {
                code += "    occludeDiffuse();\n";
            }
            if (options.occludeSpecular === SPECOCC_AO || options.occludeSpecular === SPECOCC_GLOSSDEPENDENT) {
                code += "    occludeSpecular();\n";
            }
        }

        if (options.useSpecularityFactor) {
            code += "    dSpecularLight *= dSpecularityFactor;\n";
        }

        if (options.opacityFadesSpecular === false) {
            if (options.blendType === BLEND_NORMAL || options.blendType === BLEND_PREMULTIPLIED) {
                code += "float specLum = dot((dSpecularLight + dReflection.rgb * dReflection.a), vec3( 0.2126, 0.7152, 0.0722 ));\n";
                code += "#ifdef CLEARCOAT\n specLum += dot(ccSpecularLight * ccSpecularity + ccReflection.rgb * ccReflection.a * ccSpecularity, vec3( 0.2126, 0.7152, 0.0722 ));\n#endif\n";
                code += "dAlpha = clamp(dAlpha + gammaCorrectInput(specLum), 0.0, 1.0);\n";
            }
            code += "dAlpha *= material_alphaFade;\n";
        }

        code += chunks.endPS;
        if (options.blendType === BLEND_NORMAL || options.blendType === BLEND_ADDITIVEALPHA || options.alphaToCoverage) {
            code += chunks.outputAlphaPS;
        } else if (options.blendType === BLEND_PREMULTIPLIED) {
            code += chunks.outputAlphaPremulPS;
        } else {
            code += chunks.outputAlphaOpaquePS;
        }

        if (options.msdf) {
            code += "    gl_FragColor = applyMsdf(gl_FragColor);\n";
        }

        code += "\n";
        code += end();

        if (hasPointLights) {
            code = chunks.lightDirPointPS + code;
        }
        if (usesLinearFalloff) {
            code = chunks.falloffLinearPS + code;
        }
        if (usesInvSquaredFalloff) {
            code = chunks.falloffInvSquaredPS + code;
        }
        if (usesSpot) {
            code = chunks.spotPS + code;
        }
        if (usesCookie) {
            code = chunks.cookiePS + code;
        }
        let structCode = "";
        if (code.includes("dReflection")) structCode += "vec4 dReflection;\n";
        if (code.includes("dTBN")) structCode += "mat3 dTBN;\n";
        if (code.includes("dVertexNormalW")) structCode += "vec3 dVertexNormalW;\n";
        if (code.includes("dTangentW")) structCode += "vec3 dTangentW;\n";
        if (code.includes("dBinormalW")) structCode += "vec3 dBinormalW;\n";
        if (code.includes("dViewDirW")) structCode += "vec3 dViewDirW;\n";
        if (code.includes("dReflDirW")) structCode += "vec3 dReflDirW;\n";
        if (code.includes("dHalfDirW")) structCode += "vec3 dHalfDirW;\n";
        if (code.includes("dDiffuseLight")) structCode += "vec3 dDiffuseLight;\n";
        if (code.includes("dSpecularLight")) structCode += "vec3 dSpecularLight;\n";
        if (code.includes("dLightDirNormW")) structCode += "vec3 dLightDirNormW;\n";
        if (code.includes("dLightDirW")) structCode += "vec3 dLightDirW;\n";
        if (code.includes("dLightPosW")) structCode += "vec3 dLightPosW;\n";
        if (code.includes("dShadowCoord")) structCode += "vec3 dShadowCoord;\n";
        if (code.includes("dAtten")) structCode += "float dAtten;\n";
        if (code.includes("dAttenD")) structCode += "float dAttenD;\n"; // separate diffuse attenuation for non-punctual light sources
        if (code.includes("dAtten3")) structCode += "vec3 dAtten3;\n";
        if (code.includes("dMsdf")) structCode += "vec4 dMsdf;\n";
        if (code.includes("ccReflection")) structCode += "vec4 ccReflection;\n";
        if (code.includes("ccReflDirW")) structCode += "vec3 ccReflDirW;\n";
        if (code.includes("ccSpecularLight")) structCode += "vec3 ccSpecularLight;\n";
        if (code.includes("ccSpecularityNoFres")) structCode += "float ccSpecularityNoFres;\n";

        const result = this._fsGetBeginCode() +
            this.varyings +
            this._fsGetBaseCode() +
            (options.detailModes ? chunks.detailModesPS : "") +
            structCode +
            this.frontendDecl +
            code;

        return result;
    }

    generateFragmentShader(frontendDecl, frontendCode, frontendFunc, lightingUv) {
        const options = this.options;

        this.frontendDecl = frontendDecl;
        this.frontendCode = frontendCode;
        this.frontendFunc = frontendFunc;
        this.lightingUv = lightingUv;

        if (options.pass === SHADER_PICK) {
            this.fshader = this._fsGetPickPassCode();
        } else if (options.pass === SHADER_DEPTH) {
            this.fshader = this._fsGetDepthPassCode();
        } else if (this.shadowPass) {
            this.fshader = this._fsGetShadowPassCode();
        } else if (options.customFragmentShader) {
            this.fshader = this._fsGetBeginCode() + options.customFragmentShader;
        } else {
            this.fshader = this._fsGetLitPassCode();
        }
    }

    getDefinition() {
        const result = {
            attributes: this.attributes,
            vshader: this.vshader,
            fshader: this.fshader
        };

        if (ShaderPass.isForward(this.options.pass)) {
            result.tag = SHADERTAG_MATERIAL;
        }

        return result;
    }
}

export {
    LitShader
};
